#ifndef __RCMD_SYNC__TASK_MANAGER__
#define __RCMD_SYNC__TASK_MANAGER__

#include "imho/lock_utils/simple_shared_resource.h"
#include "imho/lock_utils/condition.h"

#include "sync/task_manager.h"
#include "sync/sync_queue.h"

#include "core/sync_update_sender.h"

#include <pthread.h>
#include <unistd.h>
#include <stdint.h>

#include <string>
#include <vector>
#include <map>
#include <queue>
#include <set>
#include <tr1/memory>

class TaskManager {
private:
	SimpleSharedResource queue_lock, tasks_lock, senders_lock;
	Condition condition;

	struct key_t {
		std::string name;
		sid_t sid;
		gsid_t gsid;
		gtoid_t gtoid;
		toid_t toid;

		key_t(std::string const & _name, sid_t _sid, gsid_t _gsid, gtoid_t _gtoid, toid_t _toid) : name(_name), sid(_sid), gsid(_gsid), gtoid(_gtoid), toid(_toid) {}

        bool operator <(key_t const & key2) const {
            key_t const & key1 = *this;
            return key1.sid   <  key2.sid   || (
                   key1.sid   == key2.sid   && (
                   key1.gsid  <  key2.gsid  || (
                   key1.gsid  == key2.gsid  && (
                   key1.gtoid <  key2.gtoid || (
                   key1.gtoid == key2.gtoid && (
                   key1.toid  <  key2.toid  || (
                   key1.toid  == key2.toid  &&
                   key1.name  <  key2.name)))))));
        }
	};

	struct task_t : public key_t {
	    std::set<tid_t> blocker;
	    std::set<tid_t> blocked;
	    uint16_t runner_id;
	    uint32_t flags, running_flags;

	    template<typename It>
	    task_t(uint32_t _flags, It blocker_begin, It blocker_end, key_t const & key) : key_t(key), blocker(blocker_begin, blocker_end), blocked(), runner_id(0), flags(_flags), running_flags() {}
	};

	std::map<key_t, tid_t> task_ids;
	std::map<tid_t, task_t> tasks;
	std::queue<tid_t> queue;
	tid_t last_task_id;

	pthread_t thread;
	bool stopped;

	typedef std::map<uint16_t, std::tr1::shared_ptr<rcmdsync::SyncQueueSender> > Senders;
	Senders senders;

	SimpleSharedResource log_lock;
	int log_level;
	std::string log_file_name;
	FILE * log_file;

public:
	TaskManager() : queue_lock(), tasks_lock(), senders_lock(), condition(queue_lock), task_ids(), tasks(), queue(), last_task_id(1), thread(), stopped(true), senders(), log_lock(), log_level(), log_file_name(), log_file() {}

	static std::tr1::shared_ptr<TaskManager> task_manager;
	static TaskManager & instance();

	bool start(std::string const & log_file_name, int log_level);
	bool stop();

private:
    static void * thread_func(void *);
	int run();

private:
    bool start(tid_t tid, uint16_t runner_id, rcmdsync::SyncUpdateSender * sender);
    bool get_info(task_manager_info_t & info, rcmdsync::SyncUpdateSender * sender);

public:
	tid_t create(std::string const & name, sid_t sid, gsid_t gsid, gtoid_t gtoid, toid_t toid, std::set<tid_t> blockers, uint32_t flags);
	bool finish(tid_t tid);
	std::vector<task_manager_record_t> list();

public:
	bool subscribe(uint16_t id, std::tr1::shared_ptr<rcmdsync::SyncQueueSender> const & sender);
	bool unsubscribe(uint16_t id);

private:
	void errlog(int level, char const * format, ...) __attribute__((format(printf, 3, 4)));
};

#endif
